"""Helpers to monitor state."""
import asyncio
import logging
from collections import defaultdict
from typing import Callable, Dict, Tuple, List

from .const import MAX_REQUEST_PARAMETERS
from .utils import cyclic_tuple
from .typing import ParameterSet, SystemId, ParameterId, Parameter
from .uplink import Uplink

_LOGGER = logging.getLogger(__name__)

Callback = Callable[[SystemId, ParameterSet], None]

class Monitor():
    def __init__(self,
                 uplink: Uplink,
                 chunks: int = MAX_REQUEST_PARAMETERS):
        self._uplink = uplink
        self._callbacks = []  # type: List[Callback]
        self._parameters = defaultdict(int)  # type: Dict[Tuple[SystemId, ParameterId], int]
        self._iterator = cyclic_tuple(self._parameters.keys(), chunks)

        # start iterator up, empty list will yield empty
        next(self._iterator)

    def add_callback(self, callback):
        self._callbacks.append(callback)

    def del_callback(self, callback):
        self._callbacks.remove(callback)

    def add(self, system_id: SystemId, parameter_id: ParameterId):
        key = (system_id, parameter_id)
        self._parameters[key] += 1

    def remove(self, system_id: SystemId, parameter_id: ParameterId):
        key = (system_id, parameter_id)
        self._parameters[key] -= 1
        if not self._parameters[key]:
            del self._parameters[key]

    def postpone(self, system_id: SystemId, parameters: ParameterSet):
        self._iterator.send((system_id, parameters))

    def call_callbacks(self, system_id: SystemId, parameters: List[Parameter]):
        parameter_set = {} #  type: ParameterSet

        for parameter in parameters:
            if not parameter:
                _LOGGER.debug("Parameter not found for system %s", system_id)
                continue

            parameter_set[parameter['name']] = parameter

        for callback in self._callbacks:
            callback(system_id, parameter_set)

    async def run_once(self):
        system_id, parameter_ids = next(self._iterator)
        if not system_id:
            return

        _LOGGER.debug("Requesting: %s %s", system_id, parameter_ids)
        parameters = await asyncio.gather(*[
            self._uplink.get_parameter(system_id, parameter_id)
            for parameter_id in parameter_ids
        ])

        self.call_callbacks(system_id, parameters)

    async def run(self):
        while True:
            asyncio.sleep(4.5)
            await self.run_once()



from .const import (
    MAX_REQUEST_PARAMETERS,
    PARAM_HOTWATER_SYSTEMS,
    PARAM_CLIMATE_SYSTEMS,
    PARAM_COMPRESSOR_FREQUENCY,
    PARAM_PUMP_SPEED_HEATING_MEDIUM,
    PARAM_STATUS_COOLING,
    PARAM_VENTILATION_SYSTEMS,
    SMARTHOME_MODES,
)
from .typing import StatusItemIcon
from .types import (
    Thermostat,
    SetThermostatModel,
    VentilationSystem,
    ClimateSystem,
    HotWaterSystem,
    SmartHomeSystem,
)

from .monitor import Monitor
from .uplink import Uplink



from typing import Dict, NewType, Any, Union

SystemId = int
Identifier = Union[str, int]
ParameterId = Identifier
Parameter = Dict[str, Any]
StatusItemIcon = Dict[str, Any]
ParameterSet = Dict[ParameterId, Parameter]



"""Utilities for component."""
from itertools import islice
from typing import Iterable, Tuple, Any, Deque
from collections import deque

def cyclic_tuple(data: Iterable[Tuple[Any, Any]],
                 step: int):
    """Chunked cyclic iterator over a data set.

    Data will be returned in chunks up to `step`
    size. First tuple iterator will be used as
    grouping of chunks, and method will peek
    ahead in iterator to find `step` values
    to return.

    If `step` values are not found before hitting
    already returned value, peeking will be stopped
    """
    pending = deque()  # type: Deque[Tuple[Any, Any]]

    def postpone(pair):
        if pair in pending:
            pending.remove(pair)

    while True:
        if pending:
            curr = pending.popleft()
        else:
            curr = (None, None)

        keep = []
        grab = {curr[1]}
        while pending:
            val = pending.popleft()
            if curr[0] == val[0]:
                if val[1] in grab:
                    keep.append(val)
                    break
                grab.add(val[1])
            else:
                keep.append(val)
            if len(grab) >= step:
                break

        pending.extendleft(reversed(keep))
        postponed = yield curr[0], grab

        if len(pending) < len(data):
            pending.extend(data)

        if postponed:
            postpone(postponed)
            yield


def chunks(data, SIZE):
    it = iter(data)
    for _ in range(0, len(data), SIZE):
        yield {k: data[k] for k in islice(it, SIZE)}


def chunk_pop(data, SIZE):
    count = len(data)
    if count > SIZE:
        count = SIZE

    res = data[0:count]
    del data[0:count]
    return res



"""Exceptions used for nibeuplink"""
from typing import List, Dict, Any

ERROR_CODES = {
    -2: ('NO_CHANGE', 'Parameter value is the same as the current value.'),
    -1: ('NO_ERROR', 'Not an error, indicates success.'),
    0: ('UNKNOWN_ERROR', 'Unknown error occured.'),
    1: ('FORMAT_ERROR', 'Input format error.'),
    2: ('UNKNOWN_SYSTEM', 'Unknown system.'),
    3: ('ALREADY_CONNECTED', 'User is already connected to the system.'),
    4: ('NOT_ALLOWED_AS_SERVICE_USER', 'A service user is trying to do something he/she is not allowed to.'),
    5: ('NEEDS_RECONNECT', 'System needs a reconnect.'),
    6: ('NOT_ALLOWED_TO_RECONNECT', 'System is not allowed to be reconnected by this particular user.'),
    7: ('NOT_FOUND', 'The searched resource was not found.'),
    8: ('NOT_POSSIBLE', 'Could not accomplish the requested action.'),
    9: ('UNKNOWN_SYSTEM_UNIT', 'System unit requested was not found.'),
    10: ('UNKNOWN_HEAT_SYSTEM', 'Heat system was not found.'),
    11: ('NOT_ALLOWED_ON_SYSTEM', 'User is not allowed to do task on selected system.'),
    12: ('NO_PREMIUM_ON_SYSTEM', 'The system lacks premium subscriptions needed to access this function.'),
    13: ('TOO_MANY', 'Too many objects was requested.'),
    14: ('INVALID', 'Invalid.'),
    15: ('OUT_OF_RANGE', 'Parameter value is out of range.'),
    16: ('NOT_SETABLE', 'Parameter is not setable.'),
    17: ('NO_METADATA', 'No metadata for parameter id found.'),
    18: ('UNKNOWN_PARAMETER', 'Unknown parameter id.'),
    19: ('OAUTH_AUTHORIZATION_FAILED', 'OAuth Authorization Failed.'),
    20: ('OAUTH_INVALID_SCOPE', 'OAuth, insufficient scopes for resource.'),
    21: ('OAUTH_INVALID_ROLES', 'OAuth, user is missing role required to access resource.'),
    22: ('OAUTH_NOT_ALLOWED_CLIENT', 'OAuth, client is not allowed to access resource.'),
    23: ('UNKNOWN_CLIENT', 'Client connecting is unknown.'),
    24: ('INVALID_CLIENT_GROUP', 'Invalid client group.'),
    25: ('NOT_ALLOWED_AS_DEMO_USER', 'Not allowed as demo user.'),
    26: ('SYSTEM_OFFLINE', 'The system cannot be reached.'),
    27: ('INVALID_VOUCHER_TYPE', 'Invalid voucher type.'),
    28: ('RATE_LIMIT', 'Rate limit. Too many requests.'),
    29: ('NOT_AVAILABLE', 'Resource not available for user.'),
    30: ('INVALID_VOUCHER_SERIAL_NUMBER', 'Serial number not valid for voucher.'),
    31: ('CANT_BE_REGISTERED', 'Resource can\'t be registered.'),
    32: ('REGISTRATION_UNAVAILABLE', 'Registration is unavailable.'),
}

class UplinkException(Exception):
    pass


class UplinkResponseException(UplinkException):
    def __init__(self,
                 code: int,
                 data: Dict[str, Any]):
        self.code = code
        self.data = data
        desc = ERROR_CODES.get(code)
        if desc:
            self.message = desc[1]
            self.name = desc[0]
        else:
            self.message = 'Unknown error {}'.format(code)
            self.name = 'UNKNOWN_CODE'
        super().__init__("{}: {}\n{}".format(
            self.name,
            self.message,
            self.data))



"""Handler for uplink."""
import attr
import cattr
import logging
import asyncio
import aiohttp
import uuid
from datetime import datetime, timedelta
from typing import List, Optional, Any

from urllib.parse import urlencode, urlsplit, parse_qs

from .exceptions import UplinkResponseException, UplinkException
from .utils import chunks, chunk_pop
from .typing import (
    StatusItemIcon,
    ParameterId,
)
from .types import (
    Thermostat,
    SetThermostatModel,
)
from .const import (
    MAX_REQUEST_PARAMETERS
)

_LOGGER = logging.getLogger(__name__)


async def raise_for_status(response):
    if 400 <= response.status:
        e = aiohttp.ClientResponseError(
            response.request_info,
            response.history,
            code=response.status,
            headers=response.headers)

        if 'json' in response.headers.get('CONTENT-TYPE', ''):
            data = await response.json()
            e.message = str(data)
            raise UplinkResponseException(
                data.get('errorCode'),
                data) from e

        else:
            data = await response.text()
            raise UplinkException(data) from e

class BearerAuth(aiohttp.BasicAuth):
    def __init__(self, access_token):
        self.access_token = access_token

    def encode(self):
        return "Bearer {}".format(self.access_token)


class ParameterRequest:
    def __init__(self, parameter_id: str):
        self.parameter_id = parameter_id
        self.data         = None
        self.done         = False

class Throttle():
    """
    Throttling requests to API.

    Works by awaiting our turn then executing the request,
    and scheduling next request at a delay after the previous
    request completed.
    """
    def __init__(self, delay):
        self._delay = delay
        self._timestamp = datetime.now()

    async def __aenter__(self):
        timestamp = datetime.now()
        delay = (self._timestamp - timestamp).total_seconds()
        if delay > 0:
            _LOGGER.debug("Delaying request by %s seconds due to throttle", delay)
            await asyncio.sleep(delay)
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        self._timestamp = datetime.now() + self._delay


class Uplink():


    def __init__(self,
                 client_id,
                 client_secret,
                 redirect_uri,
                 access_data = None,
                 access_data_write = None,
                 scope = ['READSYSTEM'],
                 loop = None,
                 base = 'https://api.nibeuplink.com',
                 throttle = 4.5):

        self.redirect_uri      = redirect_uri
        self.client_id         = client_id
        self.access_data_write = access_data_write
        self.state             = None
        self.scope             = scope
        self.lock              = asyncio.Lock()
        self.throttle          = Throttle(timedelta(seconds=throttle))
        self.session           = None
        self.loop              = loop
        self.base              = base
        self.access_data       = None

        # check that the access scope is enough, otherwise ignore
        if access_data:
            if set(scope).issubset(set(access_data['scope'].split(' '))):
                self.access_data = access_data
            else:
                _LOGGER.info("Ignoring access data due to changed scope {}".format(scope))

        headers = {
            'Accept'      : 'application/json',
            'Content-Type': 'application/x-www-form-urlencoded;charset=UTF-8'
        }

        self.session           = aiohttp.ClientSession(headers   = headers,
                                                       auth      = aiohttp.BasicAuth(client_id, client_secret))
        self.requests          = {}

    async def __aenter__(self):
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await self.close()

    async def close(self):
        if self.session:
            await self.session.close()
            self.session = None

    def _handle_access_token(self, data):
        if 'access_token' not in data:
            raise ValueError('Error in reply {}'.format(data))

        if 'expires_in' in data:
            _LOGGER.debug("Token will expire in %s seconds",
                          data['expires_in'])
            expires = datetime.now() + timedelta(seconds=data['expires_in'])
        else:
            expires = None

        data['access_token_expires'] = expires.isoformat()

        self.access_data = data
        if self.access_data_write:
            self.access_data_write(data)

    async def _get_auth(self):
        if self.access_data:
            return BearerAuth(self.access_data['access_token'])
        else:
            return None

    async def get_access_token(self, code):
        payload = {
            'grant_type'   : 'authorization_code',
            'code'         : code,
            'redirect_uri' : self.redirect_uri,
        }

        async with self.session.post('{}/oauth/token'.format(self.base),
                                     data=payload) as response:
            await raise_for_status(response)
            self._handle_access_token(await response.json())

    async def refresh_access_token(self):
        if not self.access_data or 'refresh_token' not in self.access_data:
            _LOGGER.warning("No refresh token available for refresh")
            return

        _LOGGER.debug('Refreshing access token with refresh token %s',
                      self.access_data['refresh_token'])
        payload = {
            'grant_type'    : 'refresh_token',
            'refresh_token' : self.access_data['refresh_token'],
        }

        async with self.session.post('{}/oauth/token'.format(self.base),
                                     data=payload) as response:
            await raise_for_status(response)
            self._handle_access_token(await response.json())

    def get_authorize_url(self):
        self.state = uuid.uuid4().hex

        params = {
            'response_type' : 'code',
            'client_id'     : self.client_id,
            'redirect_uri'  : self.redirect_uri,
            'scope'         : ' '.join(self.scope),
            'state'         : self.state,
        }

        return '{}/oauth/authorize?{}'.format(self.base, urlencode(params))

    def get_code_from_url(self, url):
        query = parse_qs(urlsplit(url).query)
        if 'state' in query and query['state'][0] != self.state:
            raise ValueError('Invalid state in url {} expected {}'.format(query['state'], self.state))
        return query['code'][0]

    async def get(self, url, params = {}):
        async with self.lock:
            async with self.throttle:
                return await self._request(
                    self.session.get,
                    '{}/api/v1/{}'.format(self.base, url),
                    params = params,
                    headers= {},
                )

    async def put(self, url, **kwargs):
        async with self.lock:
            async with self.throttle:
                return await self._request(
                    self.session.put,
                    '{}/api/v1/{}'.format(self.base, url),
                    **kwargs
                )

    async def post(self, url, **kwargs):
        async with self.lock:
            async with self.throttle:
                return await self._request(
                    self.session.post,
                    '{}/api/v1/{}'.format(self.base, url),
                    **kwargs
                )

    async def _request(self, fun, *args, **kw):
        response = await fun(*args,
                             auth=await self._get_auth(),
                             **kw)
        try:
            if response.status == 401:
                _LOGGER.debug(response)
                _LOGGER.info("Attempting to refresh token due to error in request")
                await self.refresh_access_token()
                response.close()
                response = await fun(*args,
                                     auth=await self._get_auth(),
                                     **kw)

            await raise_for_status(response)

            if 'json' in response.headers.get('CONTENT-TYPE', ''):
                data = await response.json()
            else:
                data = await response.text()

            return data

        finally:
            response.close()

    async def get_parameter_raw(self, system_id: int, parameter_id: ParameterId):

        request = ParameterRequest(str(parameter_id))
        if system_id not in self.requests:
            self.requests[system_id] = []
        self.requests[system_id].append(request)

        # yield to any other runnable that want to add requests
        await asyncio.sleep(0)

        while True:
            async with self.lock:

                # check if we are already finished, by somebody elses request
                if request.done:
                    break

                if len(self.requests[system_id]) == 0:
                    break

                async with self.throttle:
                    # chop of as many requests from start as possible
                    requests = chunk_pop(self.requests[system_id],
                                        MAX_REQUEST_PARAMETERS)

                    _LOGGER.debug("Requesting parameters {}".format([str(x.parameter_id) for x in requests]))

                    data = await self._request(
                        self.session.get,
                        '{}/api/v1/systems/{}/parameters'.format(self.base, system_id),
                        params  = [('parameterIds', str(x.parameter_id)) for x in requests],
                        headers = {},
                    )

                lookup = {p['name']: p for p in data}

                for r in requests:
                    r.done = True
                    if r.parameter_id in lookup:
                        r.data = lookup[r.parameter_id]

        return request.data

    def add_parameter_extensions(self, data):
        if data:
            if data['displayValue'].endswith(data['unit']) and len(data['unit']):
                value = data['displayValue'][:-len(data['unit'])]

                try:
                    value = float(value)
                except ValueError:
                    pass

                data['value'] = value
            elif data['displayValue'] == '--':
                data['value'] = None
            else:
                data['value'] = data['displayValue']

    async def get_parameter(self, system_id: int, parameter_id: ParameterId):
        data = await self.get_parameter_raw(system_id, parameter_id)
        self.add_parameter_extensions(data)
        return data

    async def put_parameter(self, system_id: int, parameter_id: ParameterId, value: Any):
        headers = {
            'Accept'      : 'application/json',
            'Content-Type': 'application/json;charset=UTF-8'
        }

        data = {
            'settings': {
                str(parameter_id): value
            }
        }

        result = await self._request(
            self.session.put,
            '{}/api/v1/systems/{}/parameters'.format(self.base, system_id),
            json    = data,
            headers = headers,
        )
        return result[0]['status']

    async def get_system(self, system_id: int):
        _LOGGER.debug("Requesting system {}".format(system_id))
        return await self.get('systems/{}'.format(system_id))

    async def get_systems(self):
        _LOGGER.debug("Requesting systems")
        data = await self.get('systems')
        return data['objects']

    async def get_category_raw(self,
                               system_id: int,
                               category_id: str,
                               unit_id: int = 0):
        _LOGGER.debug("Requesting category {} on system {}".format(category_id, system_id))
        return await self.get('systems/{}/serviceinfo/categories/{}'.format(system_id, category_id),
                              {'systemUnitId': unit_id})

    async def get_category(self,
                           system_id: int,
                           category_id: str,
                           unit_id: int = 0):
        data = await self.get_category_raw(system_id, category_id, unit_id)
        for param in data:
            self.add_parameter_extensions(param)
        return data

    async def get_categories(self,
                             system_id: int,
                             parameters: bool,
                             unit_id: int = 0):
        _LOGGER.debug("Requesting categories on system {}".format(system_id))

        data = await self.get('systems/{}/serviceinfo/categories'.format(system_id),
                              {'parameters'  : str(parameters),
                               'systemUnitId': unit_id})
        for category in data:
            if category['parameters']:
                for param in category['parameters']:
                    self.add_parameter_extensions(param)
        return data

    async def get_status_raw(self, system_id: int):
        _LOGGER.debug("Requesting status on system {}".format(system_id))
        return await self.get('systems/{}/status/system'.format(system_id))

    async def get_status(self, system_id: int) -> List[StatusItemIcon]:
        data = await self.get_status_raw(system_id)
        for status in data:
            if status['parameters']:
                for param in status['parameters']:
                    self.add_parameter_extensions(param)
        return data

    async def get_units(self, system_id: int):
        _LOGGER.debug("Requesting units on system {}".format(system_id))
        return await self.get('systems/{}/units'.format(system_id))

    async def get_unit_status(self, system_id: int, unit_id: int):
        _LOGGER.debug("Requesting unit {} on system {}".format(unit_id, system_id))
        data = await self.get('systems/{}/status/systemUnit/{}'.format(system_id, unit_id))
        for status in data:
            if status['parameters']:
                for param in status['parameters']:
                    self.add_parameter_extensions(param)
        return data

    async def get_notifications(self,
                                system_id: int,
                                active: bool = True,
                                notifiction_type: str = 'ALARM'):
        _LOGGER.debug("Requesting notifications on system {}".format(system_id))
        params = {
            'active'      : str(active),
            'itemsPerPage': 100,
            'type'        : notifiction_type,
        }
        data = await self.get('systems/{}/notifications'.format(system_id), params=params)
        return data['objects']


    async def get_smarthome_mode(self,
                                 system_id: int) -> str:
        data = await self.get('systems/{}/smarthome/mode'.format(system_id))
        mode = data['mode']
        _LOGGER.debug("Get smarthome mode %s", mode)
        return mode


    async def put_smarthome_mode(self,
                                 system_id: int,
                                 mode: str) -> None:
        headers = {
            'Accept'      : 'application/json',
            'Content-Type': 'application/json;charset=UTF-8'
        }

        data = {
            'mode': mode
        }
        data = await self.put(
            'systems/{}/smarthome/mode'.format(system_id),
            json=data,
            headers=headers)
        _LOGGER.debug("Set smarthome mode %s -> %s", mode, data)


    async def get_smarthome_thermostats(self,
                                        system_id: int) -> List[Thermostat]:
        data = await self.get('systems/{}/smarthome/thermostats'.format(system_id))
        _LOGGER.debug("Get smarthome thermostats %s", data)
        return cattr.structure(data, List[Thermostat])


    async def post_smarthome_thermostats(self,
                                         system_id: int,
                                         thermostat: SetThermostatModel) -> None:
        headers = {
            'Accept'      : 'application/json',
            'Content-Type': 'application/json;charset=UTF-8'
        }

        data = attr.asdict(thermostat)
        _LOGGER.debug("Post smarthome thermostat %s -> %s", thermostat , data)
        await self.post(
            'systems/{}/smarthome/thermostats'.format(system_id),
            json=data,
            headers=headers)



from setuptools import setup, find_packages

# read the contents of your README file
from os import path
this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, 'README.rst'), encoding='utf-8') as f:
    long_description = f.read()

setup(
    name='nibeuplink',
    version='0.8.0',
    description='A python wrapper around Nibe Uplink REST API',
    long_description=long_description,
    long_description_content_type='text/x-rst',
    license='MIT',
    packages=find_packages('src'),
    package_dir = {'': 'src'},
    package_data = {
        'nibeuplink': ['py.typed'],
    },
    python_requires='>3.5',
    author='Joakim Plate',
    install_requires=[
        'asyncio',
        'aiohttp',
        'attrs>18.1',
        'cattrs',
    ],
    extras_require={
        'tests': [
            'pytest>3.6.4',
            'pytest-aiohttp',
            'pytest-cov<2.6',
            'coveralls',
            'asynctest'
        ]
    },
    entry_points = {
        'console_scripts' : ['nibeuplink=nibeuplink.__main__']
    },
    url='https://github.com/elupus/nibeuplink',
    classifiers=[
        'License :: OSI Approved :: MIT License',
        'Development Status :: 4 - Beta',
        'Environment :: Console',
        'Environment :: Plugins',
        'Framework :: AsyncIO',
    ]
)



import logging
import sys
import asyncio
import json
from urllib.parse import urldefrag, parse_qs

import nibeuplink
import argparse
import cattr

_LOGGER = logging.getLogger(__name__)

def pair(arg):
    data = arg.split('=')
    return (data[0], data[1])

def thermostat(arg):
    data = json.loads(arg)
    return cattr.structure(data, nibeuplink.SetThermostatModel)

parser = argparse.ArgumentParser(description='Read data from nibe uplink.')
parser.add_argument('--client_id'    , required=True)
parser.add_argument('--client_secret', required=True)
parser.add_argument('--redirect_uri' , required=True)

parser.add_argument('--system', type=int)
parser.add_argument('--categories', action='store_true')
parser.add_argument('--category', nargs='+')
parser.add_argument('--status', action='store_true')
parser.add_argument('--parameter', nargs='+', type=str)
parser.add_argument('--put_parameter', nargs='+', type=pair)
parser.add_argument('--units', action='store_true')
parser.add_argument('--alarms', action='store_true')
parser.add_argument('--info', action='store_true')
parser.add_argument('--unit', type=int, default=0)
parser.add_argument('--unit_status', action='store_true')
parser.add_argument('--verbose', action='store_true')
parser.add_argument('--smarthome_mode', action='store_true')
parser.add_argument('--put_smarthome_mode', type=str)
parser.add_argument('--smarthome_thermostats', action='store_true')
parser.add_argument('--post_smarthome_thermostats', type=thermostat)

args = parser.parse_args()

if args.verbose:
    root = logging.getLogger()
    root.setLevel(logging.DEBUG)

    ch = logging.StreamHandler(sys.stdout)
    ch.setLevel(logging.DEBUG)
    formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
    ch.setFormatter(formatter)
    root.addHandler(ch)


STORE = 'nibeuplink.json'


def token_read():
    try:
        with open(STORE, 'r') as myfile:
            return json.load(myfile)
    except FileNotFoundError:
        return None

def token_write(token):
    with open(STORE, 'w') as myfile:
        json.dump(token,
                  myfile,
                  indent=2)


async def run():

    scope = ['READSYSTEM']
    if args.put_parameter or args.put_smarthome_mode:
        scope.append('WRITESYSTEM')

    async with nibeuplink.Uplink(client_id         = args.client_id,
                                 client_secret     = args.client_secret,
                                 redirect_uri      = args.redirect_uri,
                                 access_data       = token_read(),
                                 access_data_write = token_write,
                                 scope             = scope) as uplink:


        if not uplink.access_data:
            auth_uri = uplink.get_authorize_url()
            print(auth_uri)
            result = input('Enter full redirect url: ')
            await uplink.get_access_token(uplink.get_code_from_url(result))

        todo = []
        if not args.system:
            todo.extend([uplink.get_systems()])
        else:

            if args.parameter:
                todo.extend([uplink.get_parameter(args.system, p) for p in args.parameter])

            if args.categories:
                todo.extend([uplink.get_categories(args.system, False, args.unit)])

            if args.category:
                todo.extend([uplink.get_category(args.system, p, args.unit) for p in args.category])

            if args.status:
                todo.extend([uplink.get_status(args.system)])

            if args.unit_status:
                todo.extend([uplink.get_unit_status(args.system, args.unit)])

            if args.units:
                todo.extend([uplink.get_units(args.system)])

            if args.alarms:
                todo.extend([uplink.get_notifications(args.system)])

            if args.info:
                todo.extend([uplink.get_notifications(args.system, notifiction_type=1)])

            if args.put_parameter:
                todo.extend([uplink.put_parameter(args.system, p[0], p[1]) for p in args.put_parameter])

            if args.smarthome_mode:
                todo.extend([uplink.get_smarthome_mode(args.system)])

            if args.put_smarthome_mode:
                todo.extend([uplink.put_smarthome_mode(args.system, args.put_smarthome_mode)])

            if args.smarthome_thermostats:
                todo.extend([uplink.get_smarthome_thermostats(args.system)])

            if args.post_smarthome_thermostats:
                todo.extend([uplink.post_smarthome_thermostats(args.system, args.post_smarthome_thermostats)])

            if not len(todo):
                todo.extend([uplink.get_system(args.system)])

        res = await asyncio.gather(*todo)
        for a in res:
            if isinstance(a, dict):
                print(json.dumps(a, indent=1))
            else:
                print(a)


if __name__ == '__main__':
    loop = asyncio.get_event_loop()
    loop.run_until_complete (run())




"""Structures representing data"""
from typing import (
    Optional,
    List,
)

import attr

from .typing import (
    ParameterId
)

@attr.s
class Parameter(object):
    parameterId = attr.ib()
    name = attr.ib()
    title = attr.ib()
    designation = attr.ib()
    unit = attr.ib()
    displayValue = attr.ib()
    rawValue = attr.ib()


@attr.s
class ParameterExtended(Parameter):
    value = attr.ib()


@attr.s
class SmartHomeSystem(object):
    name = attr.ib(type=str)


@attr.s
class Thermostat(object):
    smartHomeSystem = attr.ib(type=SmartHomeSystem)
    name = attr.ib(type=str)
    climateSystems = attr.ib(default=None, type=Optional[List[int]])
    currentTemperature = attr.ib(default=None, type=Optional[str])
    targetTemperature = attr.ib(default=None, type=Optional[str])


@attr.s
class SetThermostatModel(object):
    externalId = attr.ib(type=int)
    name = attr.ib(type=str)
    actualTemp = attr.ib(default=None, type=Optional[int]) # Multiplied by 10
    targetTemp = attr.ib(default=None, type=Optional[int]) # Multiplied by 10
    valvePosition = attr.ib(default=None, type=Optional[int])
    climateSystems = attr.ib(default=None, type=Optional[List[int]])

@attr.s
class ClimateSystem(object):
    name = attr.ib(type=str)
    return_temp = attr.ib(type=Optional[ParameterId])                 # BT3
    supply_temp = attr.ib(type=Optional[ParameterId])                 # BT2
    calc_supply_temp_heat = attr.ib(type=Optional[ParameterId])       # CSTH
    offset_cool = attr.ib(type=Optional[ParameterId])                 # OC
    room_temp = attr.ib(type=Optional[ParameterId])                   # BT50
    room_setpoint_heat = attr.ib(type=Optional[ParameterId])          # RSH
    room_setpoint_cool = attr.ib(type=Optional[ParameterId])          # RSC
    use_room_sensor = attr.ib(type=Optional[ParameterId])             # URS
    active_accessory = attr.ib(type=Optional[ParameterId])            # AA
    external_adjustment_active = attr.ib(type=Optional[ParameterId])  # EAA
    calc_supply_temp_cool = attr.ib(type=Optional[ParameterId])       # CSTC
    offset_heat = attr.ib(type=Optional[ParameterId])                 # OH
    heat_curve = attr.ib(type=Optional[ParameterId])                  # HC
    min_supply = attr.ib(type=Optional[ParameterId])                  # MIS
    max_supply = attr.ib(type=Optional[ParameterId])                  # MAS
    extra_heat_pump = attr.ib(type=Optional[ParameterId])              # EHP


@attr.s
class HotWaterSystem(object):
    name = attr.ib(type=str)
    hot_water_charging = attr.ib(type=Optional[ParameterId])            # BT6
    hot_water_top = attr.ib(type=Optional[ParameterId])                 # BT7
    hot_water_comfort_mode = attr.ib(type=Optional[ParameterId])
    hot_water_production = attr.ib(type=Optional[ParameterId])          # Active or not
    periodic_hot_water = attr.ib(type=Optional[ParameterId])
    stop_temperature_water_normal = attr.ib(type=Optional[ParameterId])
    start_temperature_water_normal = attr.ib(type=Optional[ParameterId])
    stop_temperature_water_luxary = attr.ib(type=Optional[ParameterId])
    start_temperature_water_luxary = attr.ib(type=Optional[ParameterId])
    stop_temperature_water_economy = attr.ib(type=Optional[ParameterId])
    start_temperature_water_economy = attr.ib(type=Optional[ParameterId])
    total_hot_water_compressor_time = attr.ib(type=Optional[ParameterId])
    hot_water_boost = attr.ib(type=Optional[ParameterId])


@attr.s
class VentilationSystem(object):
    name = attr.ib(type=str)
    fan_speed = attr.ib(type=Optional[ParameterId])
    exhaust_air = attr.ib(type=Optional[ParameterId])  # BT20
    extract_air = attr.ib(type=Optional[ParameterId])  # BT21
    exhaust_speed_normal = attr.ib(type=Optional[ParameterId])
    exhaust_speed_1 = attr.ib(type=Optional[ParameterId])
    exhaust_speed_2 = attr.ib(type=Optional[ParameterId])
    exhaust_speed_3 = attr.ib(type=Optional[ParameterId])
    exhaust_speed_4 = attr.ib(type=Optional[ParameterId])
    ventilation_boost = attr.ib(type=Optional[ParameterId])